@@ -2,6 +2,7 @@
 
 from __future__ import print_function
 
+from datetime import tzinfo, timedelta, datetime
 import fnmatch
 import glob
 import json
@@ -15,17 +16,32 @@ import types
 from functools import wraps
 from pprint import pprint
 
+import backoff
 import click
 import link_header
 import requests
 from requests import request
 
 
-GITHUB_API = "https://api.github.com"
-
 REQ_BUFFER_SIZE = 65536  # Chunk size when iterating a download body
 
 _github_token_cli_arg = None
+_github_api_url = None
+
+
+class _UTC(tzinfo):
+    """UTC"""
+
+    ZERO = timedelta(0)
+
+    def utcoffset(self, dt):
+        return self.ZERO
+
+    def tzname(self, dt):
+        return "UTC"
+
+    def dst(self, dt):
+        return self.ZERO
 
 
 class _NoopProgressReporter(object):
@@ -105,7 +121,7 @@ def _check_for_credentials(func):
     @wraps(func)
     def with_check_for_credentials(*args, **kwargs):
         has_github_token_env_var = "GITHUB_TOKEN" in os.environ
-        has_netrc = requests.utils.get_netrc_auth(GITHUB_API)
+        has_netrc = requests.utils.get_netrc_auth(github_api_url())
         if (not _github_token_cli_arg
                 and not has_github_token_env_var and not has_netrc):
             raise EnvironmentError(
@@ -138,7 +154,7 @@ def _progress_bar(*args, **kwargs):
 
     def formatPos(_self):
         pos = formatSize(_self.pos)
-        if _self.length_known:
+        if _self.length is not None:
             pos += '/%s' % formatSize(_self.length)
         return pos
 
@@ -162,12 +178,27 @@ def _recursive_gh_get(href, items):
         _recursive_gh_get(rels["next"], items)
 
 
+def _validate_repo_name(ctx, param, value):
+    """Callback used to check if repository argument was given."""
+    if "/" not in value:
+        raise click.BadParameter('Expected format for REPOSITORY is '
+                                 '"<org_name>/<project_name>" (e.g "jcfr/sandbox")')
+    return value
+
+
+#
+# CLI
+#
+
 @click.group()
 @click.option("--github-token", envvar='GITHUB_TOKEN', default=None,
               help="[default: GITHUB_TOKEN env. variable]")
+@click.option('--github-api-url', envvar='GITHUB_API_URL',
+              default='https://api.github.com',
+              help='[default: https://api.github.com]')
 @click.option("--progress/--no-progress", default=True,
               help="Display progress bar (default: yes).")
-def main(github_token, progress):
+def main(github_token, github_api_url, progress):
     """A CLI to easily manage GitHub releases, assets and references."""
     global progress_reporter_cls
     progress_reporter_cls.reportProgress = sys.stdout.isatty() and progress
@@ -175,10 +206,11 @@ def main(github_token, progress):
         progress_reporter_cls = _progress_bar
     global _github_token_cli_arg
     _github_token_cli_arg = github_token
+    set_github_api_url(github_api_url)
 
 
 @main.group("release")
-@click.argument('repo_name', metavar="REPOSITORY")
+@click.argument('repo_name', metavar="REPOSITORY", callback=_validate_repo_name)
 @click.pass_context
 @handle_http_error
 def gh_release(ctx, repo_name):
@@ -196,7 +228,7 @@ class AssetGroup(click.Group):
 
 
 @main.group("asset", cls=AssetGroup)
-@click.argument('repo_name', metavar="REPOSITORY")
+@click.argument('repo_name', metavar="REPOSITORY", callback=_validate_repo_name)
 @click.pass_context
 @handle_http_error
 def gh_asset(ctx, repo_name):
@@ -207,7 +239,7 @@ def gh_asset(ctx, repo_name):
 
 
 @main.group("ref")
-@click.argument('repo_name', metavar="REPOSITORY")
+@click.argument('repo_name', metavar="REPOSITORY", callback=_validate_repo_name)
 @click.pass_context
 @handle_http_error
 def gh_ref(ctx, repo_name):
@@ -217,6 +249,28 @@ def gh_ref(ctx, repo_name):
     ctx.obj = repo_name
 
 
+#
+# General
+#
+
+def github_api_url():
+    """Return GitHub API URL.
+
+    If no URL has been set, return https://api.github.com unless the
+    GITHUB_API_URL environment variable has been set.
+    """
+    if _github_api_url is None:
+        return os.environ.get('GITHUB_API_URL', 'https://api.github.com')
+    return _github_api_url
+
+
+def set_github_api_url(url):
+    """Set GitHub API URL.
+    """
+    global _github_api_url
+    _github_api_url = url
+
+
 #
 # Releases
 #
@@ -243,11 +297,24 @@ def print_release_info(release, title=None, indent=""):
         print_asset_info(i, asset, indent=indent)
 
 
+def get_release_type(release):
+    """Return the type of the release
+
+    Either 'draft', 'prerelease' (no draft) or 'release' (neither)
+    """
+    if release['draft']:
+        return 'draft'
+    if release['prerelease']:
+        return 'prerelease'
+    return 'release'
+
+
+@backoff.on_exception(backoff.expo, requests.exceptions.HTTPError, max_time=60)
 def get_releases(repo_name, verbose=False):
 
     releases = []
     _recursive_gh_get(
-        GITHUB_API + '/repos/{0}/releases'.format(repo_name), releases)
+        github_api_url() + '/repos/{0}/releases'.format(repo_name), releases)
 
     if verbose:
         list(map(print_release_info,
@@ -255,7 +322,18 @@ def get_releases(repo_name, verbose=False):
     return releases
 
 
+@backoff.on_predicate(backoff.expo, lambda x: x is None, max_time=5)
 def get_release(repo_name, tag_name):
+    """Return release
+
+    .. note::
+
+        If the release is not found (e.g the release was just created and
+        the GitHub response is not yet updated), this function is called again by
+        leveraging the `backoff` decorator.
+
+        See https://github.com/j0057/github-release/issues/67
+    """
     releases = get_releases(repo_name)
     try:
         release = next(r for r in releases if r['tag_name'] == tag_name)
@@ -362,10 +440,14 @@ def patch_release(repo_name, current_tag_name, **values):
             current_tag_name, "\n  ".join(updated)))
         print("")
 
+    if len(values.get("body", "")) >= 125000:
+        raise Exception('Failed to update release {0}. Description has {1} characters and maximum is 125000 characters'.format(
+          release["tag_name"], len(values["body"])))
+
     data.update(values)
 
     if not dry_run:
-        url = GITHUB_API + '/repos/{0}/releases/{1}'.format(
+        url = github_api_url() + '/repos/{0}/releases/{1}'.format(
             repo_name, release['id'])
         response = _request(
             'PATCH', url,
@@ -386,7 +468,7 @@ def get_assets(repo_name, tag_name, verbose=False):
         raise Exception('Release with tag_name {0} not found'.format(tag_name))
 
     assets = []
-    _recursive_gh_get(GITHUB_API + '/repos/{0}/releases/{1}/assets'.format(
+    _recursive_gh_get(github_api_url() + '/repos/{0}/releases/{1}/assets'.format(
         repo_name, release["id"]), assets)
 
     if verbose:
@@ -457,7 +539,7 @@ def gh_release_create(repo_name, tag_name, asset_pattern=None, name=None, body=N
         data["target_commitish"] = target_commitish
     if not dry_run:
         response = _request(
-              'POST', GITHUB_API + '/repos/{0}/releases'.format(repo_name),
+              'POST', github_api_url() + '/repos/{0}/releases'.format(repo_name),
               data=json.dumps(data),
               headers={'Content-Type': 'application/json'})
         response.raise_for_status()
@@ -465,7 +547,7 @@ def gh_release_create(repo_name, tag_name, asset_pattern=None, name=None, body=N
                            title="created '%s' release" % tag_name)
     else:
         print("created '%s' release (dry_run)" % tag_name)
-    if asset_pattern is not None:
+    if asset_pattern:
         gh_asset_upload(repo_name, tag_name, asset_pattern, dry_run=dry_run)
 
 
@@ -503,6 +585,8 @@ def gh_release_edit(repo_name, current_tag_name,
 @gh_release.command("delete")
 @click.argument("pattern")
 @click.option("--keep-pattern")
+@click.option("--release-type", type=click.Choice(['all', 'draft', 'prerelease', 'release']), default='all')
+@click.option("--older-than", type=int, default=0)
 @click.option("--dry-run", is_flag=True, default=False)
 @click.option("--verbose", is_flag=True, default=False)
 @click.pass_obj
@@ -512,7 +596,7 @@ def _cli_release_delete(*args, **kwargs):
 
 
 @_check_for_credentials
-def gh_release_delete(repo_name, pattern, keep_pattern=None,
+def gh_release_delete(repo_name, pattern, keep_pattern=None, release_type='all', older_than=0,
                       dry_run=False, verbose=False):
     releases = get_releases(repo_name)
     candidates = []
@@ -526,13 +610,27 @@ def gh_release_delete(repo_name, pattern, keep_pattern=None,
         if keep_pattern is not None:
             if fnmatch.fnmatch(release['tag_name'], keep_pattern):
                 continue
-        candidates.append(release['tag_name'])
-    for tag_name in candidates:
-        release = get_release(repo_name, tag_name)
+        if release_type != 'all' and release_type != get_release_type(release):
+            if verbose:
+                print('skipping release {0}: type {1} is not {2}'.format(
+                    release['tag_name'], get_release_type(release), release_type))
+            continue
+        # Assumes Zulu time.
+        # See https://stackoverflow.com/questions/127803/how-to-parse-an-iso-8601-formatted-date
+        utc = _UTC()
+        rel_date = datetime.strptime(release['created_at'], "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=utc)
+        rel_age = int((datetime.now(utc) - rel_date).total_seconds() / 60 / 60)  # In hours
+        if older_than > rel_age:
+            if verbose:
+                print('skipping release {0}: created less than {1} hours ago ({2}hrs)'.format(
+                    release['tag_name'], older_than, rel_age))
+            continue
+        candidates.append(release)
+    for release in candidates:
         print('deleting release {0}'.format(release['tag_name']))
         if dry_run:
             continue
-        url = (GITHUB_API
+        url = (github_api_url()
                + '/repos/{0}/releases/{1}'.format(repo_name, release['id']))
         response = _request('DELETE', url)
         response.raise_for_status()
@@ -583,13 +681,22 @@ def gh_release_notes(repo_name, tag_name):
     try:
         if release['body']:
             with open(filename, 'w+b') as f:
-                f.write(release['body'])
+                body = release['body']
+                if sys.version_info[0] >= 3:
+                    body = body.encode('utf-8')
+                f.write(body)
+        if 'EDITOR' not in os.environ:
+            raise EnvironmentError(
+                "This command requires editor set using EDITOR "
+                "env. variable.")
         ret = os.system('{0} {1}'.format(os.environ['EDITOR'], filename))
         if ret:
             raise Exception(
                 '{0} returned exit code {1}'.format(os.environ['EDITOR'], ret))
         with open(filename, 'rb') as f:
             body = f.read()
+            if sys.version_info[0] >= 3:
+                body = body.decode('utf-8')
         if release['body'] == body:
             return
         patch_release(repo_name, tag_name, body=body)
@@ -666,7 +773,7 @@ def _upload_release_file(
                 print("  deleting %s (invalid asset "
                       "with state set to 'new')" % asset['name'])
                 url = (
-                    GITHUB_API
+                    github_api_url()
                     + '/repos/{0}/releases/assets/{1}'.format(
                         repo_name, asset['id'])
                 )
@@ -730,7 +837,7 @@ def gh_asset_upload(repo_name, tag_name, pattern, dry_run=False, verbose=False):
     # Raise exception if no token is specified AND netrc file is found
     # BUT only api.github.com is specified. See #17
     has_github_token = "GITHUB_TOKEN" in os.environ
-    has_netrc = requests.utils.get_netrc_auth(GITHUB_API)
+    has_netrc = requests.utils.get_netrc_auth(github_api_url())
     if not has_github_token and has_netrc:
         if requests.utils.get_netrc_auth(upload_url) is None:
             raise EnvironmentError(
@@ -815,7 +922,7 @@ def gh_asset_delete(repo_name, tag_name, pattern,
         if dry_run:
             continue
         url = (
-            GITHUB_API
+            github_api_url()
             + '/repos/{0}/releases/assets/{1}'.format(repo_name, asset['id'])
         )
         response = _request('DELETE', url)
@@ -843,7 +950,7 @@ def _cli_asset_download(*args, **kwargs):
 def _download_file(repo_name, asset):
     response = _request(
         method='GET',
-        url=GITHUB_API + '/repos/{0}/releases/assets/{1}'.format(
+        url=github_api_url() + '/repos/{0}/releases/assets/{1}'.format(
             repo_name, asset['id']),
         allow_redirects=False,
         headers={'Accept': 'application/octet-stream'},
@@ -913,7 +1020,7 @@ def get_refs(repo_name, tags=None, pattern=None):
 
     refs = []
     _recursive_gh_get(
-        GITHUB_API + '/repos/{0}/git/refs'.format(repo_name), refs)
+        github_api_url() + '/repos/{0}/git/refs'.format(repo_name), refs)
 
     # If "tags" is True, keep only "refs/tags/*"
     data = refs
@@ -928,7 +1035,7 @@ def get_refs(repo_name, tags=None, pattern=None):
         try:
             tags = []
             _recursive_gh_get(
-                GITHUB_API + '/repos/{0}/git/refs/tags'.format(repo_name), tags)
+                github_api_url() + '/repos/{0}/git/refs/tags'.format(repo_name), tags)
             for ref in tags:
                 if ref["ref"] not in tag_names:
                     data.append(ref)
@@ -984,7 +1091,7 @@ def gh_ref_create(repo_name, reference, sha):
         'sha': sha
     }
     response = _request(
-          'POST', GITHUB_API + '/repos/{0}/git/refs'.format(repo_name),
+          'POST', github_api_url() + '/repos/{0}/git/refs'.format(repo_name),
           data=json.dumps(data),
           headers={'Content-Type': 'application/json'})
     response.raise_for_status()
@@ -1023,7 +1130,7 @@ def gh_ref_delete(repo_name, pattern, keep_pattern=None, tags=False,
             continue
         response = _request(
             'DELETE',
-            GITHUB_API + '/repos/{0}/git/{1}'.format(repo_name, ref['ref']))
+            github_api_url() + '/repos/{0}/git/{1}'.format(repo_name, ref['ref']))
         response.raise_for_status()
     return len(removed_refs) > 0
 
@@ -1036,7 +1143,7 @@ def gh_commit_get(repo_name, sha):
     try:
         response = _request(
             'GET',
-            GITHUB_API + '/repos/{0}/git/commits/{1}'.format(repo_name, sha))
+            github_api_url() + '/repos/{0}/git/commits/{1}'.format(repo_name, sha))
         response.raise_for_status()
         return response.json()
     except requests.exceptions.HTTPError as exc_info:
